/* Finger Relay NLM.
	This is used to relay a finger request received vi an NCP extension to
	a nominated machine.

	*/

#include <stdio.h>
#include <io.h>
#include <string.h>
#include <conio.h>
#include <time.h>

#include <advanced.h>
#include <process.h>
#include <nwmisc.h>
#include <nwconn.h>
#include <ncpext.h>
#include <nwcntask.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

BYTE FingerRelay(NCPExtensionClient *client, BYTE *requestData,
	LONG requestDataLen, BYTE *replyData, LONG *replyDataLen);

int myThreadGroupID;

char NCPExtName[] = "FINGER SERVER";

char *cpServerAddress = "146.87.3.8";
LONG lServerAddress;
struct sockaddr_in server_sockaddr;

struct queryDataStruct {
	LONG CharsEchoed;
	LONG ErrorsServicingRequests;
	LONG unused[6];
} *queryData;

long	ThreadActive;

void ShutDownNlm(void)
{
int rc;
long	lastThreadCount=0l;

if (ThreadActive) {
	ConsolePrintf("Waiting for all threads to terminated\r\n");

	while (ThreadActive){
		if (lastThreadCount!=ThreadActive) {
			ConsolePrintf("%d threads left\r\n",ThreadActive);
			lastThreadCount=ThreadActive;
			};
		ThreadSwitch();
		};
	};

rc = NWDeRegisterNCPExtension(queryData);
if (rc)
	ConsolePrintf("Error %d Deregistering NCP Extension: %s\r\n", 
							rc, NCPExtName);
else	
	ConsolePrintf("NCP Extension %s Deregistered\r\n", NCPExtName);

};

int main(void)
{
	int rc;

	myThreadGroupID = GetThreadGroupID();

	lServerAddress = inet_addr(cpServerAddress);
	if (lServerAddress == -1) {
		ConsolePrintf("%s is a malformed address\r\n", cpServerAddress);
		exit(1);
	}

	queryData->CharsEchoed = 0;
	queryData->ErrorsServicingRequests = 0;
	ThreadActive=0l;

	atexit(ShutDownNlm);

	ConsolePrintf("Registering NCP Extension: %s\r\n", NCPExtName);
	rc = NWRegisterNCPExtension(NCPExtName, FingerRelay,
			NULL, NULL, 1, 0, 0, &queryData);
	if (rc)
	{
		ConsolePrintf("Error %d registering NCP Extension: %s\r\n", rc,
				NCPExtName);
		return 2;
	}
puts("Finger Relay started\n");
SuspendThread(GetThreadID());
}


BYTE FingerRelay(NCPExtensionClient *client, BYTE *requestData,
	LONG requestDataLen, BYTE *replyData, LONG *replyDataLen)
{
	int savedThreadGroupID;
	long conn;
	struct sockaddr_in server_sockaddr;
	int	rc;
	int	s, nread;

	savedThreadGroupID = GetThreadGroupID();
	if (SetThreadGroupID(myThreadGroupID) == EFAILURE)
	{
		queryData->ErrorsServicingRequests++;
		return 0x5E;
	}

	if (DisableConnection(client->connection)) 	conn=0l;
		else 	conn=client->connection;

	ThreadActive++;

	if ( (requestDataLen >128 ) || (*replyDataLen < 128 ) )	{

		ConsolePrintf("bad parameters from client at connection: %ld, task: %ld\r\n", 
								client->connection, client->task);
		
		queryData->ErrorsServicingRequests++;
		return	TidyFinger(conn,0,savedThreadGroupID,0x5f);
	}

	server_sockaddr.sin_addr.s_addr = lServerAddress;
	server_sockaddr.sin_family = AF_INET;
	server_sockaddr.sin_port = htons(79);

	s = socket(PF_INET, SOCK_STREAM, 0);
	if (s == -1) {
		perror("socket");
		return	TidyFinger(conn,0,savedThreadGroupID,0x5f);
	}

	rc = connect(s, (struct sockaddr *) & server_sockaddr,
		     sizeof (server_sockaddr));
	if (rc == -1) {
		perror("connect");
		return	TidyFinger(conn,s,savedThreadGroupID,0x5f);
	}

	rc = write(s, requestData, requestDataLen);
	if (rc == -1) {
		perror("write");
		return	TidyFinger(conn,s,savedThreadGroupID,0x5f);
	}

	printf("Written request %s",requestData);

	/*
	 * read() does not necessarily return the full amount requested, so
	 * repeat the operation if necessary.
	 */
	
	for (nread=0;nread<*replyDataLen;nread+=rc) {
		rc = read(s, &replyData[nread], (WORD) *replyDataLen-nread);
		if (rc == -1) {
			perror("read");
			return	TidyFinger(conn,s,savedThreadGroupID,0x5f);
		}
		if (rc==0) break;
	}
	
	replyData[nread++]=0;
	*replyDataLen=nread;

	rc = close(s);
	if (rc == -1) {
		perror("close");
		return	TidyFinger(conn,0,savedThreadGroupID,0x5f);
	}

	queryData->CharsEchoed+=nread;
	return	TidyFinger(conn,0,savedThreadGroupID,0);
}

int TidyFinger(long conn,int socket,int saved, int rv)
{
if (socket) close(socket);

ThreadActive--;

if (conn) EnableConnection(conn);

if (SetThreadGroupID(saved) == EFAILURE)
		ConsolePrintf("%s reports error restoring callers thread group context\r\n",
			NCPExtName);
return rv;
};
